// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.cards.
// 
// smithers.cards is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.cards is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.cards. If not, see <http://www.gnu.org/licenses/>.

package smithers.cards;

/**
 * Class to store a playing card from a standard 52-card poker or bridge deck.
 * A card consists of a rank (Ace, 2-10, Jack, Queen or King) and a suit
 * (Clubs, Diamonds, Hearts or Spades). Instances of this class are immutable.
 */
public class Card {
    
    public static final char[] SUITS = {'C','D','H','S'};
    public static final char[] RANKS = {'A','2','3','4','5','6','7','8','9','T','J','Q','K'};
    
    private int rank; //1=ace 11=jack 12=queen 13=king.
    private char suit; //'C', 'D', 'H', 'S'.
    
    /**
     * Constructs the Ace of Clubs.
     */
    public Card() {
        rank = 1;
        suit = 'C';
    }
    
    /**
     * Creates a card with given rank and suit.
     * 
     * @param rank the rank of the card (1=A, 11=J, 12=Q, 13=K)
     * @param suit the suit of the card (C, D, H, S)
     */
    public Card(int rank, char suit) {
        if (rank < 1 || rank > 13) throw new RuntimeException("Rank is invalid.");
        if (suit != 'C' && suit != 'D' && suit != 'H' && suit != 'S') throw new RuntimeException("Suit is invalid.");
        
        this.rank = rank;
        this.suit = suit;
    }
    
    /**
     * Creates a card with given rank and suit both specified as int.
     * 
     * @param rank the rank of the card (1=A, 11=J, 12=Q, 13=K)
     * @param suit the suit of the card (0=C, 1=D, 2=H, 3=S)
     */
    public Card(int rank, int suit) {
        if (rank < 1 || rank > 13) throw new RuntimeException("Rank is invalid");
        if (suit < 0 || suit > 3) throw new RuntimeException("Suit is invalid");
        
        this.rank = rank;
        this.suit = SUITS[suit];
    }
    
    /**
     * Creates a card with given rank and suit both specified as char.
     * 
     * @param rank the rank of the card (A, 2-9, T, J, Q, K)
     * @param suit the suit of the card (C, D, H, S)
     */
    public Card(char rank, char suit) {
        switch (rank) {
            case 'A': this.rank = 1; break;
            case 'T': this.rank = 10; break;
            case 'J': this.rank = 11; break;
            case 'Q': this.rank = 12; break;
            case 'K': this.rank = 13; break;
                
            default:
                try {
                    this.rank = Integer.parseInt(String.valueOf(rank));
                } catch (NumberFormatException e) {
                    throw new RuntimeException("Rank is invalid");
                }
                if (this.rank < 1 || this.rank > 13) throw new RuntimeException("Rank is invalid");
        }
        
        if (suit != 'C' && suit != 'D' && suit != 'H' && suit != 'S') throw new RuntimeException("Suit is invalid");
        this.suit = suit;
    }
    
    /**
     * Creates a card from a single integer in the range 0-51.
     * 
     * @param index a the index of the card from 0-51
     */
    public Card(int index) {
        if (index < 0 || index > 51) {
            throw new RuntimeException("Index is invalid");
        } else if (index < 13) {
            rank = index + 1;
            suit = 'C';
        } else if (index < 26) {
            rank = index - 12;
            suit = 'D';
        } else if (index < 39) {
            rank = index - 25;
            suit = 'H';
        } else {
            rank = index - 38;
            suit = 'S';
        }
    }
    
    /**
     * Gets the rank of this card.
     * 
     * @return the rank as an {@code int}
     */
    public int getRank() {
        return rank;
    }
    
    /**
     * Gets the rank of this card as a char.
     * 
     * @return the rank as a {@code char}
     */
    public char getRankAsChar() {
        switch (rank) {
            case 1: return 'A';
            case 10: return 'T';
            case 11: return 'J';
            case 12: return 'Q';
            case 13: return 'K';
            default: return Integer.toString(rank).charAt(0);
        }
    }
    
    /**
     * Gets the rank of this card as an int.
     * 
     * @return the rank as an {@code int}
     */
    public int getRankAsInt() {
        return rank;
    }
    
    /**
     * Gets the rank of this card as a String.
     * 
     * @return a {@code String} representing the rank
     */
    public String getRankAsString() {
        switch (rank) {
            case 1: return "Ace";
            case 2: return "Two";
            case 3: return "Three";
            case 4: return "Four";
            case 5: return "Five";
            case 6: return "Six";
            case 7: return "Seven";
            case 8: return "Eight";
            case 9: return "Nine";
            case 10: return "Ten";
            case 11: return "Jack";
            case 12: return "Queen";
            default: return "King";
        }
    }
    
    /**
     * Gets the suit of this card.
     * 
     * @return the suit as a {@code char}
     */
    public char getSuit() {
        return suit;
    }
    
    /**
     * Gets the suit of this card as a char.
     * 
     * @return the suit as a {@code char}
     */
    public char getSuitAsChar() {
        return suit;
    }
    
    /**
     * Gets the suit of this card as an int.
     * 
     * @return the suit as an {@code int}
     */
    public int getSuitAsInt() {
        switch (suit) {
            case 'C': return 0;
            case 'D': return 1;
            case 'H': return 2;
        }
        return 3;
    }
    
    /**
     * Gets the suit of this card as a String.
     * 
     * @return a {@code String} representing the suit
     */
    public String getSuitAsString() {
        switch (suit) {
            case 'C': return "Clubs";
            case 'D': return "Diamonds";
            case 'H': return "Hearts";
        }
        return "Spades";
    }
    
    /**
     * Gets the colour of this card.
     * 
     * @return {@code 'B'} or {@code 'R'} depending on the suit
     */
    public char getColour() {
        if (suit == 'C' || suit == 'S') {
            return 'B';
        }
        return 'R';
    }
    
    /**
     * Gets the colour of this card.
     * 
     * @return {@code 'B'} or {@code 'R'} depending on the suit
     * @see #getColour()
     */
    public char getColor() {
        if (suit == 'C' || suit == 'S') {
            return 'B';
        }
        return 'R';
    }
    
    /**
     * Produces a 2 character String representing this card. The first
     * character is the rank and the second is the suit (e.g. "5C").
     * 
     * @return a 2 character {@code String} represention of this card
     */
    public String toString() {
        char[] temp = {this.getRankAsChar(), suit};
        return new String(temp);
    }
    
    /**
     * Produces a String representation of this card. The output is of the form
     * "<i>rank</i> of <i>suit</i>" (e.g. "Five of Clubs").
     * 
     * @return a {@code String} represention of this card
     */
    public String toStringFull() {
        return this.getRankAsString() + " of " + this.getSuitAsString();
    }
    
    /**
     * Produces an integer representation of the card based on suit then rank.
     * Gives aces the lowest value for a given suit.
     * 
     * @return an integer representing this card
     * @see #toInt(boolean)
     */
    public int toInt() {
        return toInt(false);
    }
    
    /**
     * Produces an integer representation of the card based on suit then rank.
     * Optionally gives aces the highest or lowest value in their suit.
     * 
     * @param aceHigh {@code true} if aces should be ranked high
     * @return an integer representing this card
     */
    public int toInt(boolean aceHigh) {
        int temp = rank + this.getSuitAsInt() * 13 - 1;
        if (aceHigh) {
            temp--;
            if (rank == 1) temp += 13;
        }
        return temp;
    }
    
    /**
     * Checks if another card is the same as this one. Cards are equal iff they
     * have the same suit and rank.
     * 
     * @param o the {@code Object} to test against
     * @return {@code true} if {@code o} is a {@code Card} and represents the
     *         same card as this one
     */
    public boolean equals(Object o) {
        if (!(o instanceof Card)) return false;
        Card card = (Card) o;
        return (rank == card.rank) && (suit == card.suit);
    }
    
    /**
     * Checks if another Card has the same rank as this one.
     * 
     * @param card the {@code Card} to test against
     * @return {@code true} if the cards have the same rank
     */
    public boolean sameRank(Card card) {
        return rank == card.rank;
    }
    
    /**
     * Checks if another Card has the same suit as this one.
     * 
     * @param card the {@code Card} to test against
     * @return {@code true} if the cards have the same suit
     */
    public boolean sameSuit(Card card) {
        return suit == card.suit;
    }
    
    /**
     * Checks if another Card is the same colour as this one.
     * 
     * @param card the {@code Card} to test against
     * @return {@code true} if the cards are the same colour
     */
    public boolean sameColour(Card card) {
        return (suit == 'C' || suit == 'S') ^ (card.suit == 'D' || card.suit == 'H');
    }
    
    /**
     * Checks if another Card is the same colour as this one.
     * 
     * @param card the {@code Card} to test against
     * @return {@code true} if the cards are the same colour
     * @see #sameColour(smithers.cards.Card)
     */
    public boolean sameColor(Card card) {
        return (suit == 'C' || suit == 'S') ^ (card.suit == 'D' || card.suit == 'H');
    }
    
    /**
     * Finds the signed difference between the ranks of 2 cards. The result is
     * positive if this card has the higher rank. Treats aces as low.
     * 
     * @param card {@code Card} to test against
     * @return the difference between the ranks
     */
    public int compareRank(Card card) {
        return rank - card.rank;
    }
    
}
