// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.cards.
// 
// smithers.cards is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.cards is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.cards. If not, see <http://www.gnu.org/licenses/>.

package smithers.cards;

/**
 * Represents a game of Klondike, also known as Canfield in the UK. For the
 * rules, see <a href="http://en.wikipedia.org/wiki/Klondike_(solitaire)">
 * http://en.wikipedia.org/wiki/Klondike_(solitaire)</a>. Notice that the
 * methods in this class are precisely those available to the player.
 */
public class Klondike {
    
    private CardStack[] faceDown = new CardStack[7];
    private CardStack[] tableau = new CardStack[7];
    private CardStack[] foundation = new CardStack[4];
    private CardStack deck;
    private int showWaste;
    private CardStack waste;
    private int drawsLeft;
    private int[] cardsHome = new int[4];
    private BuildingRule tableauRule = new BuildingDownByAlternatingColours(true);
    private BuildingRule foundationRule = new BuildingUpBySuits();
    
    /**
     * Creates a new game of Klondike.
     */
    public Klondike() {
        showWaste = 0;
        waste = new CardStack();
        drawsLeft = 2;
        
        for (int i=0; i<7; i++) {
            faceDown[i] = new CardStack();
            tableau[i] = new CardStack();
            if (i<4) foundation[i] = new CardStack();
        }
        
        deck = new CardStack(52);
        deck.shuffle();
        for (int i=1; i<7; i++) {
            deck.deal(faceDown[i], i);
        }
        deck.deal(tableau, -1);
        
        draw();
    }
    
    /**
     * Checks if the game has been won.
     * 
     * @return {@code true} if the game has been won
     */
    public boolean won() {
        for (int i=0; i<4; i++) {
            if (foundation[i].getLength() != 13) return false;
        }
        return true;
    }
    
    /**
     * Returns the top card in the specified foundation pile. If there are no
     * cards, this will return {@code null}.
     * 
     * @param col the index of the foundation pile
     * @return the number of facedown cards in the column
     */
    public Card foundationTopCard(int col) {
        if (col < 0 || col >= 4) throw new RuntimeException("Foundation pile does not exist.");
        if (foundation[col].getLength() == 0) return null;
        return foundation[col].topCard();
    }
    
    /**
     * Returns the number of face down cards in the specified tableau column.
     * 
     * @param col the index of the tableau column
     * @return the number of facedown cards in the column
     */
    public int faceDownLength(int col) {
        if (col < 0 || col >= 7) throw new RuntimeException("Tableau column does not exist.");
        return faceDown[col].getLength();
    }
    
    /**
     * Returns an array of the cards in the specified tableau column.
     * 
     * @param col the index of the tableau column
     * @return an array of the cards in the column
     */
    public Card[] tableauCards(int col) {
        if (col < 0 || col >= 7) throw new RuntimeException("Tableau pile does not exist.");
        return tableau[col].toArray();
    }
    
    /**
     * Returns the number of cards remaining in the deck.
     * 
     * @return the number of cards in the deck
     */
    public int deckLength() {
        return deck.getLength();
    }
    
    /**
     * Returns an array of the visible cards in the wastepile.
     * 
     * @return an array containing the (up to three) <i>visible</i> cards in
     *         the wastepile
     */
    public Card[] visibleWaste() {
        Card[] list = new Card[showWaste];
        for (int i=0; i<showWaste; i++) {
            list[i] = waste.getCard(waste.getLength() - showWaste + i);
        }
        return list;
    }
    
    /**
     * Returns the remaining number of times the wastepile can be turned over
     * to the deck.
     * 
     * @return the number of times the deck can be redealt
     */
    public int drawsLeft() {
        return drawsLeft;
    }
    
    /**
     * Draws from the deck to the wastepile.
     */
    public void draw() {
        switch (deck.getLength()) {
            case 0:
                if (drawsLeft == 0) throw new CardMoveException("No draws left.");
                showWaste = 0;
                drawsLeft -= 1;
                waste.deal(deck, 0);
                deck.deal(waste, deck.getLength() < 3 ? deck.getLength() : 3);
                showWaste = waste.getLength();
                return;
                
            case 1: case 2:
                showWaste = deck.getLength();
                deck.deal(waste, 0);
                return;
                
            default:
                showWaste = 3;
                deck.deal(waste, 3);
                return;
        }
    }
    
    /**
     * Attempts to move cards from one location to another. Cards are moved
     * from the wastepile if {@code from == 0} or the indicated column of the
     * tableau if {@code 1 <= from && from <= 7}. Cards are moved to the
     * foundations if {@code to == 0} or the indicated column of the tableau if
     * {@code 1 <= from && from <= 7}. One card will be moved, unless both
     * indices refer to the tableau, in which case whatever number of cards can
     * be moved will be. If no cards can be moved, a {@code CardMoveException}
     * will be thrown.
     * 
     * @param from the location to move from
     * @param to   the location to move to
     */
    public void move(int from, int to) {
        int count, i;
        Card fromCard;
        CardStack fromStack, toStack;
        
        if (from == 0) {
            fromStack = waste;
        } else if (from >= 1 && from <= 7) {
            fromStack = tableau[from-1];
        } else {
            throw new RuntimeException("From location does not exist.");
        }
        if (fromStack.getLength() == 0) {
            throw new CardMoveException("No cards to move.");
        }
        fromCard = fromStack.topCard();
        count = 1;
        
        if (to == 0) {
            for (i=0; i<4; i++) {
                if (foundationRule.canAddCard(foundation[i], fromCard)) break;
            }
            if (i == 4) throw new CardMoveException("This card cannot be moved to the foundations yet.");
            toStack = foundation[i];
        } else if (to >= 1 && to <= 7) {
            toStack = tableau[to-1];
            
            if (from == 0) {
                if (! tableauRule.canAddCard(toStack, fromCard)) throw new CardMoveException("This card does not go on this pile.");
            } else {
                count = tableauRule.canAddCards(toStack, fromStack);
                if (count == 0) throw new CardMoveException("This pile cannot move there.");
            }
        } else {
            throw new RuntimeException("To location does not exist.");
        }
        fromStack.move(toStack, count);
        
        if (from == 0) {
            if (showWaste > 1) showWaste -= 1;
            if (waste.getLength() == 0) showWaste = 0;
        } else if (from >= 1 && from <= 7) {
            if (tableau[from-1].getLength() == 0 && faceDown[from-1].getLength() > 0) faceDown[from-1].deal(tableau[from-1]);
        }
        return;
    }
    
}
