// Copyright 2011 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

import java.awt.Color;
import java.awt.Composite;
import java.awt.CompositeContext;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;

/**
 * Implementation of the {@code Composite} interface which provides the 16
 * standard blting operations.
 */
public class BltComposite implements Composite {
    
    /** Operation constant for {@code 0} (clear). */
    public static int OP_CLEAR = 0x0;
    /** Operation constant for {@code src & dst}. */
    public static int OP_AND = 0x1;
    /** Operation constant for {@code src & ~dst}. */
    public static int OP_AND_NOT = 0x2;
    /** Operation constant for {@code src}. */
    public static int OP_COPY = 0x3;
    /** Operation constant for {@code ~src & dst}. */
    public static int OP_NOT_AND = 0x4;
    /** Operation constant for {@code dst}. */
    public static int OP_NOOP = 0x5;
    /** Operation constant for {@code src ^ dst}. */
    public static int OP_XOR = 0x6;
    /** Operation constant for {@code src | dst}. */
    public static int OP_OR = 0x7;
    /** Operation constant for {@code ~(src | dst)}. */
    public static int OP_NOR = 0x8;
    /** Operation constant for {@code ~(src ^ dst)}. */
    public static int OP_EQUIV = 0x9;
    /** Operation constant for {@code ~dst}. */
    public static int OP_NOT = 0xa;
    /** Operation constant for {@code src | ~dst}. */
    public static int OP_OR_NOT = 0xb;
    /** Operation constant for {@code ~src}. */
    public static int OP_NOT_COPY = 0xc;
    /** Operation constant for {@code ~src | dst}. */
    public static int OP_NOT_OR = 0xd;
    /** Operation constant for {@code ~(src & dst)}. */
    public static int OP_NAND = 0xe;
    /** Operation constant for {@code 1} (set). */
    public static int OP_SET = 0xf;
    
    private int op, bandMask;
    
    /**
     * Creates a new {@code BltComposite} with a given operation.
     * 
     * @param op       the operation to perform
     * @param bandMask the band mask to apply (unmasked bands will remain
     *                 unchanged in the destination image)
     */
    public BltComposite(int op, int bandMask) {
        this.op = op & 0xf;
        this.bandMask = bandMask;
    }
    
    /**
     * Creates a context for compositing.
     * 
     * @param srcColorModel ignored
     * @param dstColorModel ignored
     * @param hints         ignored
     * @return a new context
     */
    public CompositeContext createContext(ColorModel srcColorModel, ColorModel dstColorModel, RenderingHints hints) {
        return new CompositeContext() {
            public void dispose() {}
            public void compose(Raster src, Raster dstIn, WritableRaster dstOut) {
                Rectangle rect = src.getBounds().intersection(dstIn.getBounds());
                int n = src.getNumBands();
                if (n != dstIn.getNumBands() || n != dstOut.getNumBands())
                    throw new RuntimeException("All 3 images must have the same number of bands.");
                int[] srcPix = new int[n];
                int[] dstPix = new int[n];
                for (int y=rect.y; y<rect.height; y++) for (int x=rect.x; x<rect.width; x++) {
                    src.getPixel(x, y, srcPix);
                    dstIn.getPixel(x, y, dstPix);
                    for (int k=0; k<n; k++) if ((bandMask & (1 << k)) != 0)
                        dstPix[k] = ((op & 1) != 0 ? srcPix[k] & dstPix[k] : 0) |
                                    ((op & 2) != 0 ? srcPix[k] & ~dstPix[k] : 0) |
                                    ((op & 4) != 0 ? ~srcPix[k] & dstPix[k] : 0) |
                                    ((op & 8) != 0 ? ~srcPix[k] & ~dstPix[k] : 0);
                    dstOut.setPixel(x, y, dstPix);
                }
            }
        };
    }
    
    /**
     * Tester. Loads one image file from {@code args[0]} and another from
     * {@code args[1]} and copies the former over the latter using each
     * operation of this class. Saves the results as {@code "0.png"} to {@code
     * "15.png"}.
     * 
     * @param args the arguments
     */
    public static void main(String[] args) throws IOException {
        BufferedImage src = ImageIO.read(new File(args[0]));
        BufferedImage dst = ImageIO.read(new File(args[1]));
        for (int i=0; i<0x10; i++) {
            BufferedImage image = new BufferedImage(dst.getWidth(), dst.getHeight(), dst.getType());
            Graphics2D graphics = image.createGraphics();
            graphics.setComposite(new BltComposite(3, ~0));
            graphics.drawImage(dst, null, 0, 0);
            graphics.setComposite(new BltComposite(i, 7));
            graphics.drawImage(src, null, (dst.getWidth() - src.getWidth()) / 2, (dst.getHeight() - src.getHeight()) / 2);
            ImageIO.write(image, "png", new File(i + ".png"));
        }
    }
    
}
