// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Represents a state of a mosaic puzzle.
 */
public class MosaicState implements PuzzleState<Mosaic> {
    
    Mosaic puzzle;
    int height, width;
    int[][] state;
    
    /**
     * Constructs a state for the specified puzzle.
     * 
     * @param puzzle the puzzle that this is a state of
     */
    public MosaicState(Mosaic puzzle) {
        this.puzzle = puzzle.isMutable() ? puzzle.makeImmutableCopy() : puzzle;
        height = this.puzzle.height;
        width = this.puzzle.width;
        this.state = new int[height][width];
        for (int i=0; i<height; i++) for (int j=0; j<width; j++)
            state[i][j] = -1;
    }
    
    public Mosaic getPuzzle() {
        return puzzle;
    }
    
    /**
     * Gets the state of the specified cell.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @return 0 if the cell is white, 1 if black, or -1 if unknown
     */
    public int getState(int row, int column) {
        if (row < 0 || column < 0 || row >= height || column >= width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        return state[row][column];
    }
    
    /**
     * Adds to the attempt at solving the puzzle.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @param state  1 to set coloured, 0 for not coloured, or -1 to reset
     */
    public void setState(int row, int column, int state) {
        if (row < 0 || column < 0 || row >= height || column >= width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        if (state < -1 || state > 1)
            throw new IllegalArgumentException("Invalid state: " + state + ".");
        this.state[row][column] = state;
    }
    
    /**
     * Clears the state of the puzzle.
     */
    public void clearState() {
        for (int i=0; i<height; i++) for (int j=0; j<width; j++)
            state[i][j] = -1;
    }
    
    public boolean isSolved() {
        for (int i=0; i<height; i++) for (int j=0; j<width; j++) {
            if (state[i][j] == -1) return false;
            if (puzzle.getClue(i, j) >= 0) {
                int count = 0;
                for (int k = i>0 ? i-1 : 0; k<=i+1 && k<height; k++)
                    for (int l = j>0 ? j-1 : 0; l<=j+1 && l<width; l++)
                        if (state[k][l]==1) count++;
                if (count != puzzle.getClue(i, j)) return false;
            }
        }
        return true;
    }
    
    /**
     * Generates a {@code String} representing the current situation of the
     * puzzle. Represents filled as '#', unfilled as '.' and unsure as ' '.
     * 
     * @return a {@code String} representing the state
     */
    public String toString() {
        char[] result = new char[(width + 1) * height];
        for (int i=0; i<height; i++) {
            for (int j=0; j<width; j++) {
                result[i * (width + 1) + j] = state[i][j] == 0 ? '.' : state[i][j] == 1 ? '#' : ' ';
            }
            result[i * (width + 1) + width] = '\n';
        }
        return new String(result);
    }
    
}
