// Copyright 2009, 2010, 2012 Christopher Smith
// 
// This file is part of smithers.util.
// 
// smithers.util is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.util is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.util. If not, see <http://www.gnu.org/licenses/>.

package smithers.util;

/**
 * A modified disjoint-set data structure which also partitions each set into
 * two parts. The typical usage can be described as follows. A collection of
 * objects are to be partitioned into two disjoint parts. To start with, there
 * is no information about whether any two objects are in the same part.
 * Periodically it is discovered (or decided) that two objects are, or are not,
 * in the same part. A collection of objects of this class can keep track of
 * this information.
 * 
 * @see smithers.util.UnionFindNode
 */
public class AlternatingUnionFindNode {
    
    private AlternatingUnionFindNode parent;
    private boolean samePart;
    
    /**
     * Creates a new node in a set of its own.
     */
    public AlternatingUnionFindNode() {
        parent = this;
        samePart = true;
    }
    
    private AlternatingUnionFindNode top() {
        if (parent == this) return this;
        else return parent.top();
//         AlternatingUnionFindNode oldParent = parent;
//         parent = parent.top();
//         alternates ^= oldParent.alternates;
//         return parent;
    }
    
    private boolean sTop() {
        if (parent == this) return false;
        else return ! samePart ^ parent.sTop();
    }
    
    /**
     * Tests whether this node and the given node are in the same set.
     * 
     * @param node the node to test against this one
     * @return {@code true} iff both nodes are in the same set
     */
    public boolean sameSet(AlternatingUnionFindNode node) {
        return this.top() == node.top();
    }
    
    /**
     * Tests whether this node and the given node are in the same part of a
     * set.
     * 
     * @param node the node to test against this one
     * @return {@code true} iff both nodes are in the same part of the same set
     */
    public boolean samePart(AlternatingUnionFindNode node) {
        return this.top() == node.top() && (! this.sTop() ^ node.sTop());
    }
    
    /**
     * Merges the set containing this node and the set containing the given
     * node. The two nodes given ({@code this} and {@code node}) can be placed
     * in the same part of the new set, or different parts, all other nodes in
     * the set are organised according to this. If the two nodes are already in
     * the same set, and the {@code same} parameter contradicts the existing
     * data, then the behaviour of {@code samePart(AlternatingUnionFindNode)}
     * is undefined for this set.
     * 
     * @param node a node in the set to merge with
     * @param same {@code true} if {@code this} and {@code node} should be
     *             placed in the same part of the new set
     */
    public void union(AlternatingUnionFindNode node, boolean same) {
        AlternatingUnionFindNode temp = node.top();
        temp.samePart = node.sTop() ^ this.sTop() ^ same;
        temp.parent = this.top();
    }
    
}
